/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _ARM64_MEM_EVENTS_H
#define _ARM64_MEM_EVENTS_H

extern struct perf_mem_event perf_mem_events_arm[PERF_MEM_EVENTS__MAX];

#endif /* _ARM64_MEM_EVENTS_H */
